# Django
from django.db import models, transaction
from eveuniverse.models import EveType

from skillfarm.app_settings import SKILLFARM_BULK_METHODS_BATCH_SIZE
from skillfarm.hooks import get_extension_logger
from skillfarm.providers import esi
from skillfarm.task_helper import NotModifiedError, etag_results

logger = get_extension_logger(__name__)


class CharacterSkillManager(models.Manager):
    def update_or_create_esi(self, character, force_refresh: bool = False):
        """Update or create skills for a character from ESI."""
        skills = self._fetch_data_from_esi(character, force_refresh=force_refresh)

        if not skills:
            return False

        skills_list = {
            skill["skill_id"]: skill
            for skill in skills.get("skills", [])
            if "skill_id" in skill
        }
        self._preload_types(skills_list)

        self._atomic_write(character, skills_list)
        return True

    def _preload_types(self, skills_list: dict):
        if skills_list:
            incoming_ids = set(skills_list.keys())
            existing_ids = set(self.values_list("eve_type_id", flat=True))
            new_ids = incoming_ids.difference(existing_ids)
            EveType.objects.bulk_get_or_create_esi(ids=list(new_ids))

    def _fetch_data_from_esi(self, character, force_refresh: bool = False) -> dict:
        logger.debug("%s: Fetching skills from ESI", character)

        skills_info = []
        token = character.get_token()
        try:
            skills_info_data = esi.client.Skills.get_characters_character_id_skills(
                character_id=character.character.character_id,
            )

            skills_info = etag_results(
                skills_info_data, token, force_refresh=force_refresh
            )
        except NotModifiedError:
            logger.debug(
                "No New Skillque data for: %s", character.character.character_name
            )

        return skills_info

    @transaction.atomic()
    def _atomic_write(self, character, skills_list: dict):
        incoming_ids = set(skills_list.keys())
        exiting_ids = set(
            self.filter(character=character).values_list("eve_type_id", flat=True)
        )
        obsolete_ids = exiting_ids.difference(incoming_ids)
        if obsolete_ids:
            logger.debug(
                "%s: Deleting %s obsolete skill/s", character, len(obsolete_ids)
            )
            self.filter(character=character, eve_type_id__in=obsolete_ids).delete()

        create_ids = None
        update_ids = None

        if skills_list:
            create_ids = incoming_ids.difference(exiting_ids)
            if create_ids:
                self._create_from_dict(
                    character=character, skills_list=skills_list, create_ids=create_ids
                )
            update_ids = incoming_ids.intersection(exiting_ids)
            if update_ids:
                self._update_from_dict(
                    character=character, skills_list=skills_list, update_ids=update_ids
                )

        if not obsolete_ids and not create_ids and not update_ids:
            logger.debug("%s: No changes in skills", character)

    def _create_from_dict(self, character, skills_list: dict, create_ids: set):
        logger.debug("%s: Storing %s new skills", character, len(create_ids))
        skills = [
            self.model(
                name=character.name,
                character=character,
                eve_type=EveType.objects.get(id=skill_info.get("skill_id")),
                active_skill_level=skill_info.get("active_skill_level"),
                skillpoints_in_skill=skill_info.get("skillpoints_in_skill"),
                trained_skill_level=skill_info.get("trained_skill_level"),
            )
            for skill_id, skill_info in skills_list.items()
            if skill_id in create_ids
        ]
        self.bulk_create(skills, batch_size=SKILLFARM_BULK_METHODS_BATCH_SIZE)

    def _update_from_dict(self, character, skills_list: dict, update_ids: set):
        logger.debug("%s: Updating %s skills", character, len(update_ids))
        update_pks = list(
            self.filter(character=character, eve_type_id__in=update_ids).values_list(
                "pk", flat=True
            )
        )
        skills = self.in_bulk(update_pks)
        for skill in skills.values():
            skill_info = skills_list.get(skill.eve_type_id)
            if skill_info:
                skill.active_skill_level = skill_info.get("active_skill_level")
                skill.skillpoints_in_skill = skill_info.get("skillpoints_in_skill")
                skill.trained_skill_level = skill_info.get("trained_skill_level")

        self.bulk_update(
            skills.values(),
            fields=[
                "active_skill_level",
                "skillpoints_in_skill",
                "trained_skill_level",
            ],
            batch_size=SKILLFARM_BULK_METHODS_BATCH_SIZE,
        )



"""Models for Skillfarm."""

import datetime

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveCharacter, Token

from skillfarm import app_settings
from skillfarm.hooks import get_extension_logger
from skillfarm.managers.characterskill import CharacterSkillManager
from skillfarm.managers.skillfarmaudit import SkillFarmManager
from skillfarm.managers.skillqueue import SkillqueueManager

logger = get_extension_logger(__name__)


class SkillFarmAudit(models.Model):
    """Character Audit model for app"""

    name = models.CharField(max_length=255, blank=True, null=True)

    active = models.BooleanField(default=True)

    character = models.OneToOneField(
        EveCharacter, on_delete=models.CASCADE, related_name="skillfarm_character"
    )

    notification = models.BooleanField(default=False)
    notification_sent = models.BooleanField(default=False)
    last_notification = models.DateTimeField(null=True, default=None, blank=True)

    last_update_skills = models.DateTimeField(null=True, default=None, blank=True)
    last_update_skillqueue = models.DateTimeField(null=True, default=None, blank=True)

    objects = SkillFarmManager()

    def __str__(self):
        return f"{self.character.character_name}'s Character Data"

    class Meta:
        default_permissions = ()

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            "esi-skills.read_skills.v1",
            "esi-skills.read_skillqueue.v1",
        ]

    def get_token(self) -> Token:
        """Helper method to get a valid token for a specific character with specific scopes."""
        token = (
            Token.objects.filter(character_id=self.character.character_id)
            .require_scopes(self.get_esi_scopes())
            .require_valid()
            .first()
        )
        if token:
            return token
        return False

    def skill_extraction(self) -> list[str]:
        """Check if a character has a skill extraction ready and return skill names."""
        skill_names = []
        try:
            character = SkillFarmSetup.objects.get(character=self)
        except SkillFarmSetup.DoesNotExist:
            character = None

        if character and character.skillset is not None:
            skills = CharacterSkill.objects.filter(
                character=self,
                eve_type__name__in=character.skillset,
            )

            for skill in skills:
                if skill.trained_skill_level == 5:
                    skill_names.append(skill.eve_type.name)
        return skill_names

    def skillqueue_extraction(self) -> list[str]:
        """Check if a character has a skillqueue Extraction ready and return skill names."""
        skill_names = []
        try:
            character = SkillFarmSetup.objects.get(character=self)
        except SkillFarmSetup.DoesNotExist:
            character = None

        if character and character.skillset is not None:
            skillqueue = CharacterSkillqueueEntry.objects.filter(character=self)

            for skill in skillqueue:
                if skill.is_skillqueue_ready:
                    skill_names.append(skill.eve_type.name)
        return skill_names

    def _generate_notification(self, skill_names: list[str]) -> str:
        """Generate notification for the user."""
        msg = _("%(charname)s: %(skillname)s") % {
            "charname": self.character.character_name,
            "skillname": ", ".join(skill_names),
        }
        return msg

    def get_finished_skills(self) -> list[str]:
        """Return a list of finished skills."""
        skill_names = set()
        skills = self.skill_extraction()
        skillqueue = self.skillqueue_extraction()

        if skillqueue:
            skill_names.update(skillqueue)
        if skills:
            skill_names.update(skills)

        return list(skill_names)

    @property
    def is_active(self):
        time_ref = timezone.now() - datetime.timedelta(
            days=app_settings.SKILLFARM_CHAR_MAX_INACTIVE_DAYS
        )
        try:
            is_active = True

            is_active = self.last_update_skillqueue > time_ref
            is_active = self.last_update_skills > time_ref

            if self.active != is_active:
                self.active = is_active
                self.save()

            return is_active
        except Exception:  # pylint: disable=broad-exception-caught
            return False

    @property
    def is_cooldown(self) -> bool:
        """Check if a character has a notification cooldown."""
        if self.last_notification is None:
            return False
        return True


class SkillFarmSetup(models.Model):
    """Skillfarm Character Skill Setup model for app"""

    id = models.AutoField(primary_key=True)

    name = models.CharField(max_length=255, blank=True, null=True)

    character = models.OneToOneField(
        "SkillFarmAudit", on_delete=models.CASCADE, related_name="skillfarm_setup"
    )

    skillset = models.JSONField(default=dict, blank=True, null=True)

    def __str__(self):
        return f"{self.skillset}'s Skill Setup"

    objects = SkillFarmManager()

    class Meta:
        default_permissions = ()


class CharacterSkill(models.Model):
    """Skillfarm Character Skill model for app"""

    name = models.CharField(max_length=255, blank=True, null=True)

    character = models.ForeignKey(
        "SkillFarmAudit", on_delete=models.CASCADE, related_name="character_skills"
    )
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    active_skill_level = models.PositiveIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)]
    )
    skillpoints_in_skill = models.PositiveBigIntegerField()
    trained_skill_level = models.PositiveBigIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)]
    )

    objects = CharacterSkillManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character}-{self.eve_type.name}"

    @property
    def is_exc_ready(self) -> bool:
        """Check if skill extraction is ready."""
        try:
            character = SkillFarmSetup.objects.get(character=self.character)
        except SkillFarmSetup.DoesNotExist:
            character = None

        if character and character.skillset is not None:
            skills = CharacterSkill.objects.filter(
                character=self.character,
                eve_type__name__in=character.skillset,
            )
            for skill in skills:
                if skill.trained_skill_level == 5:
                    return True
        return False


class CharacterSkillqueueEntry(models.Model):
    """Skillfarm Skillqueue model for app"""

    name = models.CharField(max_length=255, blank=True, null=True)

    character = models.ForeignKey(
        "SkillFarmAudit",
        on_delete=models.CASCADE,
        related_name="skillqueue",
    )

    queue_position = models.PositiveIntegerField(db_index=True)
    finish_date = models.DateTimeField(default=None, null=True)
    finished_level = models.PositiveIntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(5)]
    )
    level_end_sp = models.PositiveIntegerField(default=None, null=True)
    level_start_sp = models.PositiveIntegerField(default=None, null=True)
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    start_date = models.DateTimeField(default=None, null=True)
    training_start_sp = models.PositiveIntegerField(default=None, null=True)

    # TODO: Add to Notification System
    has_no_skillqueue = models.BooleanField(default=False)
    last_check = models.DateTimeField(default=None, null=True)

    objects = SkillqueueManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character}-{self.queue_position}"

    @property
    def is_active(self) -> bool:
        """Returns true when this skill is currently being trained"""
        return bool(self.finish_date) and self.finish_date > self.start_date

    @property
    def is_skillqueue_ready(self) -> bool:
        """Check if skill from skillqueue is rdy from end date."""
        if (
            self.is_active
            and self.finish_date <= timezone.now()
            and self.finished_level == 5
        ):
            return True
        return False



from ninja import NinjaAPI
from ninja.security import django_auth

from django.conf import settings

from skillfarm.api import character
from skillfarm.hooks import get_extension_logger

logger = get_extension_logger(__name__)

api = NinjaAPI(
    title="Geuthur API",
    version="0.2.0",
    urls_namespace="skillfarm:api",
    auth=django_auth,
    csrf=True,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)

# Add the character endpoints
character.setup(api)



"""Model for Permission."""

from django.db import models

from skillfarm.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class General(models.Model):
    """General model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app, Skillfarm."),
            ("corp_access", "Has access to all characters in the corporation."),
            ("admin_access", "Has access to all characters."),
        )



"""PvE Views"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as trans
from django.views.decorators.http import require_POST
from esi.decorators import token_required
from eveuniverse.models import EveMarketPrice

from allianceauth.authentication.models import UserProfile
from allianceauth.eveonline.models import EveCharacter

from skillfarm import forms
from skillfarm.api.helpers import get_character
from skillfarm.hooks import get_extension_logger
from skillfarm.models.skillfarm import SkillFarmAudit, SkillFarmSetup

logger = get_extension_logger(__name__)


# pylint: disable=unused-argument
def add_info_to_context(request, context: dict) -> dict:
    """Add additional information to the context for the view."""
    theme = None
    try:
        user = UserProfile.objects.get(id=request.user.id)
        theme = user.theme
    except UserProfile.DoesNotExist:
        pass

    new_context = {
        **{"theme": theme},
        **context,
    }
    return new_context


@login_required
@permission_required("skillfarm.basic_access")
def index(request):
    """Index View"""
    return redirect(
        "skillfarm:skillfarm", request.user.profile.main_character.character_id
    )


@login_required
@permission_required("skillfarm.basic_access")
def skillfarm(request, character_id=None):
    """Main Skillfarm View"""
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    context = {
        "page_title": "Skillfarm",
        "character_id": character_id,
        "forms": {
            "confirm": forms.ConfirmForm(),
            "skillset": forms.SkillSetForm(),
        },
    }
    context = add_info_to_context(request, context)
    return render(request, "skillfarm/skillfarm.html", context=context)


@login_required
@permission_required("skillfarm.basic_access")
def character_overview(request, character_id=None):
    """Character Overview"""
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    context = {
        "page_title": "Character Admin",
        "character_id": character_id,
    }
    context = add_info_to_context(request, context)

    return render(request, "skillfarm/overview.html", context=context)


@login_required
@token_required(scopes=SkillFarmAudit.get_esi_scopes())
@permission_required("skillfarm.basic_access")
def add_char(request, token):
    """Add Character to Skillfarm"""
    # pylint: disable=import-outside-toplevel
    from skillfarm.tasks import update_character_skillfarm

    try:
        character = EveCharacter.objects.get_character_by_id(token.character_id)
        char, _ = SkillFarmAudit.objects.update_or_create(
            character=character, defaults={"name": token.character_name}
        )
        update_character_skillfarm.apply_async(
            args=[char.character.character_id], kwargs={"force_refresh": True}
        )
    except SkillFarmAudit.DoesNotExist:
        msg = trans("Character not found")
        messages.error(request, msg)
        return redirect("skillfarm:index")

    msg = trans("{character_name} successfully added to Skillfarm System").format(
        character_name=char.character.character_name,
    )
    messages.success(request, msg)
    return redirect("skillfarm:index")


@login_required
@permission_required("skillfarm.basic_access")
@require_POST
def remove_char(request, character_id: list):
    """Remove Character from Skillfarm"""
    # Retrieve character_id from GET parameters
    character_id = int(request.POST.get("character_id", 0))

    # Check Permission
    perm, _ = get_character(request, character_id)

    if not perm:
        msg = trans("Permission Denied")
        messages.error(request, msg)
        return redirect("skillfarm:skillfarm", character_id=character_id)

    try:
        character = SkillFarmAudit.objects.get(character__character_id=character_id)
        character.delete()
    except SkillFarmAudit.DoesNotExist:
        msg = trans("Character/s not found")
        messages.error(request, msg)
        return redirect("skillfarm:skillfarm", character_id=character_id)

    msg = trans("{character_name} successfully Deleted").format(
        character_name=character.character.character_name,
    )
    messages.success(request, msg)

    return redirect("skillfarm:skillfarm", character_id=character_id)


@login_required
@permission_required("skillfarm.basic_access")
@require_POST
def switch_alarm(request, character_id: list):
    """Switch Character Notification Alarm"""
    # Check Permission
    perm, __ = get_character(request, character_id)
    form = forms.ConfirmForm(request.POST)
    if form.is_valid():
        if not perm:
            msg = trans("Permission Denied")
            return JsonResponse(
                {"success": False, "message": msg}, status=403, safe=False
            )

        character_id = form.cleaned_data["character_id"]
        characters = [character_id]

        try:
            characters = SkillFarmAudit.objects.filter(
                character__character_id__in=characters
            )
            if characters:
                for c in characters:
                    c.notification = not c.notification
                    c.save()
            else:
                raise SkillFarmAudit.DoesNotExist
            msg = trans("Alarm/s successfully updated")
        except SkillFarmAudit.DoesNotExist:
            msg = "Character/s not found"
            return JsonResponse(
                {"success": False, "message": msg}, status=404, safe=False
            )
    else:
        msg = "Invalid Form"
    return JsonResponse({"success": True, "message": msg}, status=200, safe=False)


@login_required
@permission_required("skillfarm.basic_access")
@require_POST
def skillset(request, character_id: list):
    """Edit Character SkillSet"""
    # Check Permission
    perm, __ = get_character(request, character_id)
    form = forms.SkillSetForm(request.POST)

    if form.is_valid():
        if not perm:
            msg = trans("Permission Denied")
            return JsonResponse(
                {"success": False, "message": msg}, status=403, safe=False
            )
        character_id = form.cleaned_data["character_id"]
        selected_skills = form.cleaned_data["selected_skills"]
        try:
            skillset_list = selected_skills.split(",") if selected_skills else None
            character = SkillFarmAudit.objects.get(character__character_id=character_id)
            SkillFarmSetup.objects.update_or_create(
                character=character, defaults={"skillset": skillset_list}
            )
        except SkillFarmAudit.DoesNotExist:
            msg = trans("Character not found")
            return JsonResponse(
                {"success": False, "message": msg}, status=404, safe=False
            )

        msg = trans("{character_name} Skillset successfully updated").format(
            character_name=character.character.character_name,
        )
    else:
        msg = "Invalid Form"
        return JsonResponse({"success": False, "message": msg}, status=400, safe=False)
    return JsonResponse({"success": True, "message": msg}, status=200, safe=False)


@login_required
@permission_required("skillfarm.basic_access")
def skillfarm_calc(request, character_id=None):
    """Skillfarm Calc View"""
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    skillfarm_dict = {}
    error = False
    try:
        plex = EveMarketPrice.objects.get(eve_type_id=44992)
        injector = EveMarketPrice.objects.get(eve_type_id=40520)
        extractor = EveMarketPrice.objects.get(eve_type_id=40519)

        month = plex.average_price * 500
        month12 = plex.average_price * 300
        month24 = plex.average_price * 275

        monthcalc = (injector.average_price * 3.5) - (
            month + (extractor.average_price * 3.5)
        )
        month12calc = (injector.average_price * 3.5) - (
            month12 + (extractor.average_price * 3.5)
        )
        month24calc = (injector.average_price * 3.5) - (
            month24 + (extractor.average_price * 3.5)
        )

        skillfarm_dict["plex"] = plex
        skillfarm_dict["injektor"] = injector
        skillfarm_dict["extratkor"] = extractor

        skillfarm_dict["calc"] = {
            "month": monthcalc,
            "month12": month12calc,
            "month24": month24calc,
        }
    except EveMarketPrice.DoesNotExist:
        EveMarketPrice.objects.update_from_esi()
        error = True

    context = {
        "error": error,
        "character_id": character_id,
        "page_title": "Skillfarm Calc",
        "skillfarm": skillfarm_dict,
    }

    return render(request, "skillfarm/calculator.html", context=context)



"""App Tasks"""

import datetime

# Third Party
# pylint: disable=no-name-in-module
from celery import shared_task

from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from allianceauth.notifications import notify
from allianceauth.services.tasks import QueueOnce

from skillfarm.app_settings import (
    SKILLFARM_NOTIFICATION_COOLDOWN,
    SKILLFARM_STALE_STATUS,
)
from skillfarm.decorators import when_esi_is_available
from skillfarm.hooks import get_extension_logger
from skillfarm.models.skillfarm import (
    CharacterSkill,
    CharacterSkillqueueEntry,
    SkillFarmAudit,
)
from skillfarm.task_helper import enqueue_next_task, no_fail_chain

logger = get_extension_logger(__name__)


@shared_task
@when_esi_is_available
def update_all_skillfarm(runs: int = 0):
    characters = SkillFarmAudit.objects.select_related("character").all()
    for character in characters:
        update_character_skillfarm.apply_async(args=[character.character.character_id])
        runs = runs + 1
    logger.info("Queued %s Skillfarm Updates", runs)


@shared_task(bind=True, base=QueueOnce)
def update_character_skillfarm(
    self, character_id, force_refresh=False
):  # pylint: disable=unused-argument
    character = SkillFarmAudit.objects.get(character__character_id=character_id)
    skip_date = timezone.now() - datetime.timedelta(hours=SKILLFARM_STALE_STATUS)
    que = []
    mindt = timezone.now() - datetime.timedelta(days=7)
    logger.debug(
        "Processing Audit Updates for %s", format(character.character.character_name)
    )
    if (character.last_update_skillqueue or mindt) <= skip_date or force_refresh:
        que.append(update_char_skillqueue.si(character_id, force_refresh=force_refresh))

    if (character.last_update_skills or mindt) <= skip_date or force_refresh:
        que.append(update_char_skills.si(character_id, force_refresh=force_refresh))

    enqueue_next_task(que)

    logger.debug("Queued %s Tasks for %s", len(que), character.character.character_name)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_skillqueue",
)
@no_fail_chain
def update_char_skillqueue(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    character = SkillFarmAudit.objects.get(character__character_id=character_id)
    CharacterSkillqueueEntry.objects.update_or_create_esi(
        character, force_refresh=force_refresh
    )
    character.last_update_skillqueue = timezone.now()
    character.save()


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_skills",
)
@no_fail_chain
def update_char_skills(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    character = SkillFarmAudit.objects.get(character__character_id=character_id)
    CharacterSkill.objects.update_or_create_esi(character, force_refresh=force_refresh)
    character.last_update_skills = timezone.now()
    character.save()


# pylint: disable=unused-argument, too-many-locals
@shared_task(bind=True, base=QueueOnce)
def check_skillfarm_notifications(self, runs: int = 0):
    characters = SkillFarmAudit.objects.select_related("character").all()
    warnings = {}
    notified_characters = []

    # Create a dictionary to map main characters to their alts
    main_to_alts = {}
    for character in characters:
        main_character = (
            character.character.character_ownership.user.profile.main_character
        )
        if main_character not in main_to_alts:
            main_to_alts[main_character] = []
        main_to_alts[main_character].append(character)

    for main_character, alts in main_to_alts.items():
        msg_items = []
        for alt in alts:
            if alt.notification and not alt.is_cooldown:
                skill_names = alt.get_finished_skills()
                if skill_names:
                    # Create and Add Notification Message
                    msg = alt._generate_notification(skill_names)
                    msg_items.append(msg)
                    notified_characters.append(alt)
        if msg_items:
            # Add each message to Main Character
            warnings[main_character] = "\n".join(msg_items)
    if warnings:
        for main_character, msg in warnings.items():
            logger.debug(
                "Skilltraining has been finished for %s Skills: %s",
                main_character.character_name,
                msg,
            )
            title = _("Skillfarm Notifications")
            full_message = format_html(
                "Following Skills have finished training: \n{}", msg
            )
            notify(
                title=title,
                message=full_message,
                user=main_character.character_ownership.user,
                level="warning",
            )

            # Set notification_sent to True for all characters that were notified
            for character in notified_characters:
                character.notification_sent = True
                character.last_notification = timezone.now()
                character.save()

            runs = runs + 1

    # Reset notification for characters that have not been notified for more than a day
    for character in characters:
        if (
            character.last_notification is not None
            and character.last_notification
            < timezone.now() - datetime.timedelta(days=SKILLFARM_NOTIFICATION_COOLDOWN)
        ):
            logger.debug(
                "Notification Reseted for %s",
                character.character.character_name,
            )
            character.last_notification = None
            character.notification_sent = False
            character.save()

    logger.info("Queued %s Skillfarm Notifications", runs)



from datetime import datetime
from typing import Any, List, Optional

from ninja import Schema


class Message(Schema):
    message: str


class SkillFarm(Schema):
    character_id: int
    character_name: str
    corporation_id: int
    corporation_name: str
    active: Optional[bool]
    notification: Optional[bool]
    last_update: Optional[datetime]
    skillset: Any
    skills: Any
    skill_names: Any
    is_active: Optional[bool]
    extraction_ready: Optional[bool]


class SkillFarmFilter(Schema):
    characters: List[Any]
    skills: List[Any]
